"""Generate the focal fields from a Gaussian beam in a pupil and visualize the results."""

import matplotlib.pyplot as plt
import numpy as np

from leb.just_focus import InputField, Polarization, Pupil, Stop


def main(plot=True) -> None:
    mesh_size = 64
    pupil = Pupil(
        na=1.4,
        refractive_index=1.518,
        wavelength_um=0.561,
        mesh_size=mesh_size,
        stop=Stop.TANH,
    )

    inputs = InputField.gaussian_pupil(
        beam_center=(0.0, 0.0),
        waist=1.0,
        mesh_size=mesh_size,
        polarization=Polarization.LINEAR_Y,
    )

    results = pupil.propgate(0.0, inputs, padding_factor=4)

    _, axs = plt.subplots(1, 3, figsize=(12, 6))
    axs[0].imshow(
        np.abs(results.field_y),
        origin="lower",
        extent=(results.x_um[0], results.x_um[-1], results.y_um[0], results.y_um[-1]),
    )
    axs[1].imshow(
       np.angle(results.field_y),
       origin="lower",
       extent=(results.x_um[0], results.x_um[-1], results.y_um[0], results.y_um[-1]),
    )
    axs[2].plot(results.x_um, np.angle(results.field_y)[mesh_size // 2, :])
    axs[2].set_title("Phase, line profile through center")
    axs[2].set_xlabel(r"x, $\mu m$")
    
    if plot:
        plt.show()


if __name__ == "__main__":
    np.seterr("raise")
    main()



"""Electromagnetic fields in the focus of a high NA microscope objective."""
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

from .dtypes import Complex, Float


@dataclass(frozen=True)
class FocalField:
    field_x: NDArray[Complex]
    field_y: NDArray[Complex]
    field_z: NDArray[Complex]
    x_um: NDArray[Float]
    y_um: NDArray[Float]

    def intensity(self, normalize: bool = True) -> NDArray[Float]:
        I = np.abs(self.field_x)**2 + np.abs(self.field_y)**2 + np.abs(self.field_z)**2
        if normalize:
            return I / np.max(I)
        return I



import numpy as np

from leb.just_focus import Pupil


def test_pad_width():
    arr = np.array([[1.0, 2.0], [3.0, 4.0]])
    padding_factor = 2
    expected_padding = (3, 3)  # Add 3 elements to each side to get 2 * 2**2 = 8

    result = Pupil._pad_width(arr.shape, padding_factor=padding_factor)

    assert result[0] == expected_padding
    assert result[1] == expected_padding



import numpy as np

from leb.just_focus import Float, FocalField


def test_focal_field_intensity_return_type_is_float():
    field_x = np.random.uniform(-1, 1, size=(4, 4)) + 1j * np.random.uniform(-1, 1, size=(4, 4))
    field_y, field_z = field_x.copy(), field_x.copy()
    x_um, y_um = np.meshgrid([0, 1, 2, 3], [0, 1, 2, 3])

    focal_field = FocalField(field_x, field_y, field_z, x_um, y_um)
    intensity = focal_field.intensity()

    assert intensity.dtype == Float



import numpy as np

from leb.just_focus import InputField, Polarization, Pupil
from leb.just_focus.plots import plot_inputs


def test_smoke() -> None:
    """Smoke test to ensure everything runs without errors."""
    np.seterr("raise")

    mesh_size = 32
    pupil = Pupil(
        na=1.4,
        refractive_index=1.518,
        wavelength_um=0.561,
        mesh_size=mesh_size
    )
    inputs = InputField.uniform_pupil(mesh_size, Polarization.CIRCULAR_LEFT)

    results = pupil.propgate(0.0, inputs, padding_factor=5)

    results.intensity()

    plot_inputs(inputs, pupil, show=False)



"""Input fields for the propagation algorithm."""

from __future__ import annotations
from dataclasses import dataclass
from enum import StrEnum

import numpy as np
from numpy.typing import NDArray

from .dtypes import Complex, Float


class Polarization(StrEnum):
    LINEAR_X = "linear_x"
    LINEAR_Y = "linear_y"
    CIRCULAR_LEFT = "circular_left"
    CIRCULAR_RIGHT = "circular_right"

    def arrays(self, mesh_size: int) -> tuple[NDArray[Complex], NDArray[Complex]]:
        match self:
            case Polarization.LINEAR_X:
                polarization_x = np.ones((mesh_size, mesh_size), dtype=Complex)
                polarization_y = np.zeros((mesh_size, mesh_size), dtype=Complex)
            case Polarization.LINEAR_Y:
                polarization_x = np.zeros((mesh_size, mesh_size), dtype=Complex)
                polarization_y = np.ones((mesh_size, mesh_size), dtype=Complex)
            case Polarization.CIRCULAR_LEFT:
                polarization_x = np.ones((mesh_size, mesh_size), dtype=Complex) / np.sqrt(2)
                polarization_y = 1j * np.ones((mesh_size, mesh_size), dtype=Complex) / np.sqrt(2)
            case Polarization.CIRCULAR_RIGHT:
                polarization_x = np.ones((mesh_size, mesh_size), dtype=Complex) / np.sqrt(2)
                polarization_y = -1j * np.ones((mesh_size, mesh_size), dtype=Complex) / np.sqrt(2)
            case _:
                raise ValueError(f"Unsupported polarization: {polarization}")
        
        return polarization_x, polarization_y


class HalfmoonPhase(StrEnum):
    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"
    PLUS_45 = "plus_45"
    MINUS_45 = "minus_45"

    def arrays(
        self,
        mesh_size: int,
        phase: float = np.pi,
        phase_mask_center: tuple[float, float] = (0.0, 0.0)
    ) -> tuple[NDArray[Float], NDArray[Float]]:
        normed_coords = np.linspace(-1, 1, mesh_size)
        x, y = np.meshgrid(normed_coords, normed_coords)
        x0, y0 = phase_mask_center
        x -= x0
        y -= y0

        phase_x = np.zeros((mesh_size, mesh_size), dtype=Float)
        match self:
            case HalfmoonPhase.HORIZONTAL:
                phase_x[x >= 0] = phase
            case HalfmoonPhase.VERTICAL:
                phase_x[y >= 0] = phase
            case HalfmoonPhase.PLUS_45:
                phase_x[(x + y) >= 0] = phase
            case HalfmoonPhase.MINUS_45:
                phase_x[(x - y) >= 0] = phase

        phase_y = phase_x.copy()

        return phase_x, phase_y



@dataclass
class InputField:
    """Factory class for creating input fields for the pupil.

    Each direction may be specified independently, which models separate beam shaping
    elements for the x- and y-directions. In many common cases, the amplitudes and
    phases will be the same in both x- and y-directions and only the polarization will
    differ.
    
    Attributes
    ----------
    amplitude_x : NDArray[Float]
        The amplitude of the field for the x-direction.
    amplitude_y : NDArray[Float]
        The amplitude of the field for the y-direction.
    phase_x : NDArray[Float]
        The phase of the field for the x-direction.
    phase_y : NDArray[Float]
        The phase of the field for the y-direction.
    polarization_x : NDArray[Complex]
        The polarization state of the field for the x-direction.
    polarization_y : NDArray[Complex]
        The polarization state of the field for the y-direction.

    Methods
    -------
    gaussian_pupil(beam_center, waist, mesh_size, polarization)
        Create a Gaussian pupil field with a specified waist size.
    uniform_pupil(mesh_size, polarization)
        Create a uniform pupil field with specified polarization.

    """
    amplitude_x: NDArray[Float]
    amplitude_y: NDArray[Float]
    phase_x: NDArray[Float]
    phase_y: NDArray[Float]
    polarization_x: NDArray[Complex]
    polarization_y: NDArray[Complex]

    @staticmethod
    def _gaussian_amplitude(
        beam_center: tuple[float, float],
        waist: float | tuple[float, float],
        mesh_size: int
    ) -> tuple[NDArray[Float], NDArray[Float]]:
        """Calculate a Gaussian amplitude for the pupil field."""
        if isinstance(waist, (int, float)):
            waist_x = waist_y = waist
        else:
            waist_x, waist_y = waist

        normed_coords = np.linspace(-1, 1, mesh_size)
        x, y = np.meshgrid(normed_coords, normed_coords)
        x0: float = beam_center[0]
        y0: float = beam_center[1]
        amplitude_x = np.exp(-(x - x0)**2 / waist_x**2 - (y - y0)**2 / waist_y**2)
        amplitude_y = np.copy(amplitude_x)

        return amplitude_x, amplitude_y

    @classmethod
    def gaussian_pupil(
        cls,
        beam_center: tuple[float, float],
        waist: float | tuple[float, float],
        mesh_size: int,
        polarization: Polarization
    ) -> InputField:
        """Create a Gaussian pupil field with a specified waist size.

        Parameters
        ----------
        beam_center : tuple of float
            The center of the Gaussian beam in normalized pupil coordinates (x, y).
        waist : float or tuple of float
            The waist size of the Gaussian beam in normalized pupil coordinates. If a
            single float is provided, it is used for both x and y dimensions.
        mesh_size : int
            The size of the mesh grid for the pupil field.
        polarization : Polarization
            The polarization state of the field.

        Returns
        -------
        InputField
            The input field with Gaussian amplitude and specified polarization.

        """
        polarization_x, polarization_y = polarization.arrays(mesh_size)
        amplitude_x, amplitude_y = cls._gaussian_amplitude(beam_center, waist, mesh_size)

        phase_x = np.zeros((mesh_size, mesh_size), dtype=Float)
        phase_y = np.zeros((mesh_size, mesh_size), dtype=Float)

        return InputField(
            amplitude_x=amplitude_x,
            amplitude_y=amplitude_y,
            phase_x=phase_x,
            phase_y=phase_y,
            polarization_x=polarization_x,
            polarization_y=polarization_y,
        )
    
    @classmethod
    def gaussian_halfmoon_pupil(
        cls,
        beam_center: tuple[float, float],
        waist: float | tuple[float, float],
        mesh_size: int,
        polarization:Polarization,
        orientation: HalfmoonPhase = HalfmoonPhase.HORIZONTAL,
        phase: float = np.pi,
        phase_mask_center: tuple[float, float] = (0.0, 0.0)
    ) -> InputField:
        polarization_x, polarization_y = polarization.arrays(mesh_size)
        amplitude_x, amplitude_y = cls._gaussian_amplitude(beam_center, waist, mesh_size)

        phase_x, phase_y = orientation.arrays(mesh_size, phase, phase_mask_center)

        return InputField(
            amplitude_x=amplitude_x,
            amplitude_y=amplitude_y,
            phase_x=phase_x,
            phase_y=phase_y,
            polarization_x=polarization_x,
            polarization_y=polarization_y,
        )

    @classmethod
    def uniform_pupil(cls, mesh_size: int, polarization: Polarization) -> InputField:
        polarization_x, polarization_y = polarization.arrays(mesh_size)
            
        amplitude_x = np.ones((mesh_size, mesh_size), dtype=Float)
        amplitude_y = np.ones((mesh_size, mesh_size), dtype=Float)
        phase_x = np.zeros((mesh_size, mesh_size), dtype=Float)
        phase_y = np.zeros((mesh_size, mesh_size), dtype=Float)

        return InputField(
            amplitude_x=amplitude_x,
            amplitude_y=amplitude_y,
            phase_x=phase_x,
            phase_y=phase_y,
            polarization_x=polarization_x,
            polarization_y=polarization_y,
        )



from __future__ import annotations
from dataclasses import dataclass, field
from enum import StrEnum

import numpy as np
from numpy.fft import ifft2, ifftshift
from numpy.typing import NDArray

from .dtypes import Float
from .inputs import InputField
from .focal_fields import FocalField


class Stop(StrEnum):
    """Defines the type of stop used in the pupil.

    Attributes
    ----------
    UNIFORM : str
        A uniform stop, where the pupil is filled completely.
    TANH : str
        A stop defined by a hyperbolic tangent function. This helps to improve accuracy
        in the focal field by gradually reducing the amplitude near the stop's edge.

        See Leutenegger, et al., Opt. Express 14, 11277-91 (2006), eq. 16 for details.
    """
    UNIFORM = "uniform"
    TANH = "tanh"

    def array(self, px: NDArray[Float], py: NDArray[Float]) -> NDArray[Float]:
        match self:
            case Stop.UNIFORM:
                return ((px**2 + py**2) <= 1).astype(Float)
            case Stop.TANH:
                mesh_size = px.shape[0]
                return 0.5 * (1 + np.tanh(1.5 * mesh_size * (1 - np.sqrt(px**2 + py**2))))
            case _:
                raise ValueError(f"Unsupported stop type: {self}")


@dataclass
class Pupil:
    na: float = 1.4
    wavelength_um: float = 0.532
    refractive_index: float = 1.518
    focal_length_mm: float = 3.3333
    mesh_size: int = 64
    stop: Stop = Stop.UNIFORM

    x_mm: NDArray[Float] = field(init=False, repr=False)
    y_mm: NDArray[Float] = field(init=False, repr=False)
    stop_arr: NDArray[Float] = field(init=False, repr=False)
    stop_radius_mm: float = field(init=False, repr=False)
    kx: NDArray[Float] = field(init=False, repr=False)
    ky: NDArray[Float] = field(init=False, repr=False)
    kz: NDArray[Float] = field(init=False, repr=False)
    k: float = field(init=False, repr=False)

    def __post_init__(self) -> None:
        normed_coords = np.linspace(-1, 1, self.mesh_size)
        
        px, py = np.meshgrid(normed_coords, normed_coords)
        self.stop_arr = self.stop.array(px, py)

        # Far field coordinate system
        f = self.focal_length_mm / 1e3 # Convert focal length from mm to meters
        x_scaling = f * self.na
        xinf = px * x_scaling
        yinf = py * x_scaling

        # Save pupil coordinates to make plotting easier
        self.x_mm = normed_coords * x_scaling * 1e3
        self.y_mm = normed_coords * x_scaling * 1e3
        self.stop_radius_mm = x_scaling * 1e3

        # Angular spectrum coordinate system
        k0 = 2 * np.pi * 1e6 / self.wavelength_um # Convert wavelength from um to meters
        self.k: float = k0 * self.refractive_index
        self.kx, self.ky = k0 * xinf / f, k0 * yinf / f

        # Set values of kz outside the pupil to 1 to avoid division by zero later
        self.kz = np.sqrt(np.maximum(1, self.k**2 - self.kx**2 - self.ky**2))

    def propgate(self, z_um: float, inputs: InputField, padding_factor: int = 2) -> FocalField:
        """Propagate the input field to the focal plane at distance z.
        
        Parameters
        ----------
        z_um : float
            The distance to propagate the field in micrometers.
        inputs : InputField
            The input field to propagate.
        padding_factor : int, optional
            The factor by which to pad the input field arrays before propagation. To
            maintain array sizes that are powers of 2, arrays will be padded so that
            their padded shapes are `2**padding_factor * arr.shape[0]` and
            `2**padding_factor * arr.shape[1]`. Default is 2.
            
        Returns
        -------
        FocalField
            The field in the desired z plane.

        """
        z = z_um * 1e-6  # Convert z from micrometers to meters
        defocus = np.exp(1j * self.kz * z)
        kz_root = np.sqrt(self.kz)
        k_transverse_sq = self.kx**2 + self.ky**2

        far_field_x = defocus * self.stop_arr * (
            inputs.polarization_x * inputs.amplitude_x * np.exp(1j * inputs.phase_x) * (self.ky**2 + self.kx**2 * self.kz / self.k) + \
            inputs.polarization_y * inputs.amplitude_y * np.exp(1j * inputs.phase_y) * (-self.kx * self.ky + self.kx * self.ky * self.kz / self.k)
        ) / k_transverse_sq / kz_root
        far_field_y = defocus * self.stop_arr * (
            inputs.polarization_x * inputs.amplitude_x * np.exp(1j * inputs.phase_x) * (-self.kx * self.ky + self.kx * self.ky * self.kz / self.k) + \
            inputs.polarization_y * inputs.amplitude_y * np.exp(1j * inputs.phase_y) * (self.kx**2 + self.ky**2 * self.kz / self.k)
        ) / k_transverse_sq / kz_root
        far_field_z = defocus * self.stop_arr * (
            inputs.polarization_x * inputs.amplitude_x * np.exp(1j * inputs.phase_x) * (-k_transverse_sq * self.kx / self.k) + \
            inputs.polarization_y * inputs.amplitude_y * np.exp(1j * inputs.phase_y) * (-k_transverse_sq * self.ky / self.k)
        ) / k_transverse_sq / kz_root

        padding: tuple[tuple[int, int], tuple[int, int]] = self._pad_width(far_field_x.shape, padding_factor)
        far_field_x_padded = np.pad(far_field_x, padding, mode='constant', constant_values=0)
        far_field_y_padded = np.pad(far_field_y, padding, mode='constant', constant_values=0)
        far_field_z_padded = np.pad(far_field_z, padding, mode='constant', constant_values=0)

        # Compute the phase correction due to not sampling the origin
        # See Herrera and Quinto-Su, arxiv:2211.06725 (2022), section
        # Optical System: Mesh and note the typo in the first minus sign (it should be
        # a plus sign).
        auxillary_mesh_size = self.mesh_size * 2**padding_factor
        auxillary_coords = np.linspace(-auxillary_mesh_size // 2, auxillary_mesh_size // 2, auxillary_mesh_size)
        px, py = np.meshgrid(auxillary_coords, auxillary_coords)
        correction_scaling = 1j * 2 * np.pi * 0.5 / px.shape[0]
        phase_correction = np.exp(correction_scaling * px + correction_scaling * py)

        field_x = ifftshift(ifft2(ifftshift(far_field_x_padded))) * phase_correction
        field_y = ifftshift(ifft2(ifftshift(far_field_y_padded))) * phase_correction
        field_z = ifftshift(ifft2(ifftshift(far_field_z_padded))) * phase_correction

        dx = self.wavelength_um / 2 / self.na / 2**padding_factor
        dy = dx
        x_um = dx * np.linspace(-field_x.shape[0] // 2, field_x.shape[0] // 2 - 1, field_x.shape[0])
        y_um = dy * np.linspace(-field_y.shape[1] // 2, field_y.shape[1] // 2 - 1, field_y.shape[1])

        return FocalField(field_x=field_x, field_y=field_y, field_z=field_z, x_um=x_um, y_um=y_um)
    
    @staticmethod
    def _pad_width(array_shape: tuple[int, int], padding_factor: int) -> tuple[tuple[int, int], tuple[int, int]]:
        """Calculate the padding width for an array."""
        padded_shape = (2**padding_factor * array_shape[0], 2**padding_factor * array_shape[1])
        pad_height = (padded_shape[0] - array_shape[0]) // 2
        pad_width = (padded_shape[1] - array_shape[1]) // 2
        return ((pad_height, pad_height), (pad_width, pad_width))



"""Generate a half-moon pupil with an off center Gaussian beam and visualize the results."""
import matplotlib.pyplot as plt
from matplotlib.patches import Circle
import numpy as np

from leb.just_focus import HalfmoonPhase, InputField, Polarization, Pupil, Stop


def main(plot=True) -> None:
    mesh_size = 64
    pupil = Pupil(
        na=1.4,
        refractive_index=1.518,
        wavelength_um=0.561,
        mesh_size=mesh_size,
        stop=Stop.TANH,
    )

    inputs = InputField.gaussian_halfmoon_pupil(
        beam_center=(0.0, 0.5),
        waist=2.0,
        mesh_size=mesh_size,
        polarization=Polarization.LINEAR_Y,
        orientation=HalfmoonPhase.MINUS_45,
        phase=np.pi,
        phase_mask_center=(0.0, 0.0),
    )

    results = pupil.propgate(0.0, inputs, padding_factor=4)

    _, axs = plt.subplots(2, 4, figsize=(12, 6))
    axs[0, 0].imshow(
        inputs.amplitude_x,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 0].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 0].set_ylabel("y, mm")
    axs[0, 0].set_title("Amplitude, x")

    axs[0, 1].imshow(
        inputs.amplitude_y,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 1].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 1].set_title("Amplitude, y")

    axs[0, 2].imshow(
        inputs.phase_x,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 2].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 2].set_title("Phase, x")

    axs[0, 3].imshow(
        inputs.phase_y,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 3].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 3].set_title("Phase, y")

    axs[1, 0].imshow(
        np.abs(inputs.polarization_x),
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[1, 0].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[1, 0].set_title("Polarization, x")
    axs[1, 0].set_xlabel("x, mm")
    axs[1, 0].set_ylabel("y, mm")

    axs[1, 1].imshow(
        np.abs(inputs.polarization_y),
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[1, 1].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[1, 1].set_title("Polarization, y")
    axs[1, 1].set_xlabel("x, mm")

    axs[1, 2].imshow(pupil.stop_arr, vmin=0, vmax=1)
    axs[1, 2].set_title("Stop")
    axs[1, 2].set_xlabel("x, mm")


    axs[1, 3].imshow(
        results.intensity(normalize=True),
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(results.x_um[0], results.x_um[-1], results.y_um[0], results.y_um[-1])
    )
    axs[1, 3].set_title("Intensity")
    axs[1, 3].set_xlabel(r"x, $\mu m$")
    axs[1, 3].set_xlim(-1, 1)
    axs[1, 3].set_ylim(-1, 1)
    
    if plot:
        plt.show()


if __name__ == "__main__":
    np.seterr("raise")
    main()



"""Tools for plotting the inputs to a simulation.

Requires the "plot" extra to be installed.
"""

import matplotlib.pyplot as plt
from matplotlib.patches import Circle
import numpy as np

from leb.just_focus import InputField, Pupil

def plot_inputs(inputs: InputField, pupil: Pupil, show: bool = True) -> None:
    _, axs = plt.subplots(2, 4, figsize=(12, 6))
    axs[0, 0].imshow(
        inputs.amplitude_x,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 0].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 0].set_ylabel("y, mm")
    axs[0, 0].set_title("Amplitude, x")

    axs[0, 1].imshow(
        inputs.amplitude_y,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 1].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 1].set_title("Amplitude, y")

    axs[0, 2].imshow(
        inputs.phase_x,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 2].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 2].set_title("Phase, x")

    axs[0, 3].imshow(
        inputs.phase_y,
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[0, 3].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[0, 3].set_title("Phase, y")

    axs[1, 0].imshow(
        np.abs(inputs.polarization_x),
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[1, 0].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[1, 0].set_title("Polarization, x")
    axs[1, 0].set_xlabel("x, mm")
    axs[1, 0].set_ylabel("y, mm")

    axs[1, 1].imshow(
        np.abs(inputs.polarization_y),
        vmin=0,
        vmax=1,
        origin="lower",
        extent=(pupil.x_mm[0], pupil.x_mm[-1], pupil.y_mm[0], pupil.y_mm[-1]),
    )
    axs[1, 1].add_artist(Circle((0, 0), radius=pupil.stop_radius_mm, color='k', fill=False, linewidth=2))
    axs[1, 1].set_title("Polarization, y")
    axs[1, 1].set_xlabel("x, mm")

    axs[1, 2].imshow(pupil.stop_arr, vmin=0, vmax=1)
    axs[1, 2].set_title("Stop")
    axs[1, 2].set_xlabel("x, mm")

    axs[1, 3].remove()  # Remove the empty subplot

    if show:
        plt.show()
